"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AzureDevOpsAgileService = void 0;
const azure_devops_node_api_1 = require("azure-devops-node-api");
const CoreApi_1 = require("azure-devops-node-api/CoreApi");
const WorkItemTrackingInterfaces_1 = require("azure-devops-node-api/interfaces/WorkItemTrackingInterfaces");
const WorkApi_1 = require("azure-devops-node-api/WorkApi");
const WorkItemTrackingApi_1 = require("azure-devops-node-api/WorkItemTrackingApi");
const constants_1 = require("../../../../constants");
const utils_1 = require("../../../../utils");
const baseAgileService_1 = require("../../baseAgileService");
const azureDevOpsFieldName_1 = require("./azureDevOpsFieldName");
const azureDevOpsUtils_1 = require("./azureDevOpsUtils");
class AzureDevOpsAgileService extends baseAgileService_1.BaseAgileService {
    constructor(config) {
        super(config);
        // Projects
        this.createProject = () => __awaiter(this, void 0, void 0, function* () {
            throw new Error("Not implemented");
        });
        // Backlog Items
        this.getBacklogItems = (ids) => __awaiter(this, void 0, void 0, function* () {
            const workItems = yield this.workItemTracking.getWorkItems(ids.map((id) => +id), undefined, undefined, WorkItemTrackingInterfaces_1.WorkItemExpand.Relations);
            return yield Promise.all(workItems.map((workItem) => this.mapWorkItem(workItem, true)));
        });
        this.createBacklogItems = (items, parent) => __awaiter(this, void 0, void 0, function* () {
            const backlogItems = [];
            for (const item of items) {
                // Make AzDO API call
                const workItem = yield this.workItemTracking.createWorkItem(undefined, azureDevOpsUtils_1.AzureDevOpsUtils.createPatchDocument(item, parent === null || parent === void 0 ? void 0 : parent.url), this.projectName, azureDevOpsUtils_1.AzureDevOpsUtils.getWorkItemType(item.type));
                // Map work item from AzDO model to BacklogItem
                const createdBacklogItem = yield this.mapWorkItem(workItem);
                // Create children if applicable
                if (item.children) {
                    createdBacklogItem.children = yield this.createBacklogItems(item.children, createdBacklogItem);
                }
                // Add created backlog item to list
                backlogItems.push(createdBacklogItem);
            }
            return backlogItems;
        });
        this.deleteBacklogItems = (ids) => __awaiter(this, void 0, void 0, function* () {
            yield Promise.all(ids.map((id) => this.workItemTracking.deleteWorkItem(+id, this.projectName)));
        });
        // Sprints
        this.getSprint = (id) => __awaiter(this, void 0, void 0, function* () {
            const teamContext = yield this.getTeamContext();
            const teamIteration = yield utils_1.retryAsync(() => this.workApi.getTeamIteration(teamContext, id));
            if (!teamIteration) {
                throw new Error(`Could not retrieve sprint ${id}`);
            }
            const { name, attributes } = teamIteration;
            return {
                id,
                name: name || "",
                startDate: attributes === null || attributes === void 0 ? void 0 : attributes.startDate,
                finishDate: attributes === null || attributes === void 0 ? void 0 : attributes.finishDate,
            };
        });
        this.createProviderSprints = (sprints) => __awaiter(this, void 0, void 0, function* () {
            const teamContext = yield this.getTeamContext();
            utils_1.Logger.log(`Creating ${sprints.length} sprints`);
            const createdSprints = [];
            for (const sprint of sprints) {
                createdSprints.push(yield this.createProviderSprint(sprint, teamContext));
            }
            return sprints;
        });
        this.deleteSprint = (id) => __awaiter(this, void 0, void 0, function* () {
            const teamContext = yield this.getTeamContext();
            yield this.workApi.deleteTeamIteration(teamContext, id);
        });
        const { providerOptions } = config;
        const projectName = utils_1.Config.getValueWithDefault(constants_1.ConfigValue.AzDOProjectName, providerOptions.projectName);
        if (!projectName) {
            throw new Error("Need to configure an AzDO project name in config file or environment variable");
        }
        this.projectName = projectName;
        const personalAccessToken = utils_1.Config.getValueWithDefault(constants_1.ConfigValue.AzDOAccessToken, providerOptions.personalAccessToken);
        if (!personalAccessToken) {
            throw new Error("Need to configure an AzDO personal access token in config file or environment variable");
        }
        const baseUrl = utils_1.Config.getValueWithDefault(constants_1.ConfigValue.AzDOBaseUrl, providerOptions.baseUrl);
        if (!baseUrl) {
            throw new Error("Need to configure an AzDO base URL in config file or environment variable");
        }
        // Instantiate auth handler to be used in API clients
        const authHandler = azure_devops_node_api_1.getPersonalAccessTokenHandler(personalAccessToken);
        // Instantiate various API clients
        this.coreApi = new CoreApi_1.CoreApi(baseUrl, [authHandler]);
        this.workItemTracking = new WorkItemTrackingApi_1.WorkItemTrackingApi(baseUrl, [authHandler]);
        this.workApi = new WorkApi_1.WorkApi(baseUrl, [authHandler]);
    }
    // Private functions
    createProviderSprint(sprint, teamContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, startDate, finishDate } = sprint;
            // Creates classification node and returns identifier required for sprint ID
            const identifier = yield this.createOrUpdateClassificationNode(sprint, teamContext);
            // Create new iteration
            const result = yield this.workApi.postTeamIteration({
                id: identifier,
                name,
                attributes: {
                    startDate,
                    finishDate,
                },
            }, teamContext);
            // Assign generated ID from Azure DevOps to sprint
            sprint.id = result.id;
            utils_1.Logger.log(`Sprint '${name}' created`);
            return sprint;
        });
    }
    createOrUpdateClassificationNode(sprint, teamContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, startDate, finishDate } = sprint;
            utils_1.Guard.empty(name);
            yield this.deleteNodeAndIterationIfExists(name, teamContext);
            const { identifier } = yield this.workItemTracking.createOrUpdateClassificationNode({
                name,
                attributes: {
                    startDate: startDate === null || startDate === void 0 ? void 0 : startDate.toISOString(),
                    finishDate: finishDate === null || finishDate === void 0 ? void 0 : finishDate.toISOString(),
                },
            }, this.projectName, WorkItemTrackingInterfaces_1.TreeStructureGroup.Iterations);
            if (!identifier) {
                throw new Error(`Was not able to retrieve identifier for ${this.projectName}`);
            }
            return identifier;
        });
    }
    deleteNodeAndIterationIfExists(name, teamContext) {
        return __awaiter(this, void 0, void 0, function* () {
            const node = yield this.workItemTracking.getClassificationNode(this.projectName, WorkItemTrackingInterfaces_1.TreeStructureGroup.Iterations, name);
            if (node && node.identifier) {
                yield this.workApi.deleteTeamIteration(teamContext, node.identifier);
                yield this.workItemTracking.deleteClassificationNode(this.projectName, WorkItemTrackingInterfaces_1.TreeStructureGroup.Iterations, name);
            }
        });
    }
    getTeamContext() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.teamContext) {
                return this.teamContext;
            }
            const { name, id, defaultTeam } = yield this.coreApi.getProject(this.projectName);
            this.teamContext = {
                project: name,
                projectId: id,
                team: defaultTeam === null || defaultTeam === void 0 ? void 0 : defaultTeam.name,
                teamId: defaultTeam === null || defaultTeam === void 0 ? void 0 : defaultTeam.id,
            };
            return this.teamContext;
        });
    }
    mapWorkItem(workItem, includeChildren = false) {
        return __awaiter(this, void 0, void 0, function* () {
            const { id, fields, url } = workItem;
            const workItemType = fields[azureDevOpsFieldName_1.AzureDevOpsFieldName.workItemType];
            return {
                id: id.toString(),
                name: fields[azureDevOpsFieldName_1.AzureDevOpsFieldName.title],
                description: fields[azureDevOpsFieldName_1.AzureDevOpsFieldName.description],
                acceptanceCriteria: fields[azureDevOpsFieldName_1.AzureDevOpsFieldName.acceptanceCriteria],
                type: azureDevOpsUtils_1.AzureDevOpsUtils.getBacklogItemType(workItemType),
                children: includeChildren ? yield this.getBacklogItemChildren(workItem) : undefined,
                url,
            };
        });
    }
    getBacklogItemChildren(workItem) {
        return __awaiter(this, void 0, void 0, function* () {
            const { relations } = workItem;
            if (relations) {
                const childIds = [];
                for (const relation of relations) {
                    const { attributes, url } = relation;
                    if (!attributes || !(attributes.name === "Child")) {
                        continue;
                    }
                    const id = url === null || url === void 0 ? void 0 : url.substr(url.lastIndexOf("/") + 1);
                    if (id) {
                        childIds.push(id);
                    }
                }
                if (childIds.length) {
                    return yield this.getBacklogItems(childIds);
                }
            }
        });
    }
}
exports.AzureDevOpsAgileService = AzureDevOpsAgileService;
